# coding: utf-8
"""
CALS Builder
============

This module can construct a CALS table from
an instance of type :class:`~benker.table.Table`.

Specifications and examples:

- The CALS DTD is available online in the OASIS web site:
  `CALS Table Model Document Type Definition <https://www.oasis-open.org/specs/a502.htm>`_.

- An example of CALS table is available in Wikipedia:
  `CALS Table Model <https://en.wikipedia.org/wiki/CALS_Table_Model>`_

"""
import collections
import itertools
import re

from lxml import etree

from benker.builders.base_builder import BaseBuilder
from benker.units import convert_value


def _get_border_style(styles, style):
    parts = styles.get(style, "")
    parts = parts.split(" ") if parts else []
    value = None
    for part in parts:
        if not part.endswith("pt") and part != "auto" and not part.startswith("#"):
            value = part
    return value


def _get_frame_attr(styles):
    frame_styles = {'border-top': False, 'border-right': False, 'border-bottom': False, 'border-left': False}
    for style in frame_styles:
        value = _get_border_style(styles, style) or u"none"
        frame_styles[style] = value != u"none"
    top = frame_styles['border-top']
    bottom = frame_styles['border-bottom']
    left = frame_styles['border-left']
    right = frame_styles['border-right']
    return {
        (True, True, True, True): u"all",
        (True, True, False, False): u"topbot",
        (False, False, True, True): u"sides",
        (True, False, False, False): u"top",
        (False, True, False, False): u"bottom",
    }.get((top, bottom, left, right), u"none")


def _get_colsep_attr(styles, style="x-cell-border-right"):
    value = _get_border_style(styles, style)
    return None if value is None else "0" if value == "none" else "1"


def _get_rowsep_attr(styles, style="x-cell-border-bottom"):
    value = _get_border_style(styles, style)
    return None if value is None else "0" if value == "none" else "1"


def revision_mark(name, attrs):
    target = etree.tounicode(etree.Element(name, attrib=attrs))
    target = target.replace("<{0} ".format(name), "").replace("/>", "")
    rev_pi = etree.ProcessingInstruction(name, target)
    return rev_pi


class CalsBuilder(BaseBuilder):
    """
    CALS table builder.
    """

    def __init__(self, width_unit="mm", table_in_tgroup=False, **options):
        """
        Initialize the builder.

        :param str width_unit:
            Unit to use for column widths.
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.

        :param bool table_in_tgroup:
            Where should we put the table properties:
            - ``False`` to put the properties in the ``<table>`` element,
            - ``True`` to put the properties in the ``<tgroup>`` element.

        :keyword options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.
        """
        # Internal state of the table used during building
        self._table = None
        self._table_colsep = u"0"
        self._table_rowsep = u"0"
        # options
        self.width_unit = width_unit
        self.table_in_tgroup = table_in_tgroup
        super(CalsBuilder, self).__init__(**options)

    def generate_table_tree(self, table):
        """
        Build the XML table from the Table instance.

        :type  table: benker.table.Table
        :param table: Table

        :return: Table tree
        """
        table_elem = self.build_table(table)
        return table_elem

    def build_table(self, table):
        """
        Build the CALS ``<table>`` element.

        CALS attributes:

        -   ``@colsep`` is built from the "x-cell-border-right" style.
            Default value is "0" (not displayed).

        -   ``@rowsep`` is built from the "x-cell-border-bottom" style.
            Default value is "0" (not displayed).

        -   ``@tabstyle`` is built from the table nature.

        -   ``@orient`` is built from the "x-sect-orient" style (orientation of the current section).
            Possible values are "port" (portrait, the default) or "land" (landscape).

        -   ``@pgwide`` is built from the "x-sect-cols" style (column number of the current section).
            Default value is "0" (width of the current column).

        .. note::

           ``@colsep``, ``@rowsep`` and ``@tabstyle`` attributes are generated only
           if the *table_in_tgroup* options is ``False``.

        .. attention::

           According to the `CALS specification <https://www.oasis-open.org/specs/tm9502.html#c37ab3>`_,
           the default value for ``@colsep`` and ``@rowsep`` should be "1".
           But, having this value as a default is really problematic for conversions:
           most of nowadays formats, like Office Open XML and CSS, consider that the default
           value is "no border" (a.k.a: ``border: none``).
           So, setting "0" as a default value is a better choice.

        :type  table: benker.table.Table
        :param table: Table

        :return: The newly-created ``<table>`` element.
        """
        self._table = table
        self._table_colsep = u"0"
        self._table_rowsep = u"0"
        table_styles = table.styles
        attrs = {'frame': _get_frame_attr(table_styles), }
        if not self.table_in_tgroup:
            self._table_colsep = attrs['colsep'] = _get_colsep_attr(table_styles) or "0"
            self._table_rowsep = attrs['rowsep'] = _get_rowsep_attr(table_styles) or "0"
            if table.nature is not None:
                attrs['tabstyle'] = table.nature
        if 'x-sect-orient' in table_styles:
            attrs['orient'] = {"landscape": "land", "portrait": "port"}[table_styles['x-sect-orient']]
        if 'x-sect-cols' in table_styles:
            attrs['pgwide'] = "1" if table_styles['x-sect-cols'] == "1" else "0"
        table_elem = etree.Element(u"table", attrib=attrs)
        self.build_tgroup(table_elem, table)
        return table_elem

    def build_tgroup(self, table_elem, table):
        """
        Build the CALS ``<tgroup>`` element.

        CALS attributes:

        -   ``@cols`` is the total number of columns.

        -   ``@colsep`` is built from the "x-cell-border-right" style.
            Default value is "0" (not displayed).

        -   ``@rowsep`` is built from the "x-cell-border-bottom" style.
            Default value is "0" (not displayed).

        -   ``@tgroupstyle`` is built from the table nature.

        .. note::

           ``@colsep``, ``@rowsep`` and ``@tgroupstyle`` attributes are generated only
           if the *table_in_tgroup* options is ``True``.

        :type  table_elem: etree._Element
        :param table_elem: Parent element: ``<table>``.

        :type  table: benker.table.Table
        :param table: Table

        :return: The newly-created ``<tgroup>`` element.
        """
        table_styles = table.styles
        attrs = {u'cols': str(len(table.cols))}
        if self.table_in_tgroup:
            self._table_colsep = attrs['colsep'] = _get_colsep_attr(table_styles) or "0"
            self._table_rowsep = attrs['rowsep'] = _get_rowsep_attr(table_styles) or "0"
            if table.nature is not None:
                attrs['tgroupstyle'] = table.nature
        group_elem = etree.SubElement(table_elem, u"tgroup", attrib=attrs)
        for col in table.cols:
            self.build_colspec(group_elem, col)
        # -- group rows by head/body/foot
        groups = [(k, list(g)) for k, g in itertools.groupby(table.rows, key=lambda r: r.nature)]
        # -- sort the groups in the order: head => foot => body
        groups = sorted(groups, key=lambda item: {"head": 0, "foot": 1, "body": 2}[item[0]])
        group_tags = {"head": u"thead", "body": u"tbody", "foot": u"tfoot"}
        for nature, row_list in groups:
            nature_tag = group_tags[nature]
            self.build_tbody(group_elem, row_list, nature_tag)

    # noinspection PyMethodMayBeStatic
    def build_colspec(self, group_elem, col):
        """
        Build the CALS ``<colspec>`` element.

        CALS attributes:

        -   ``@colname`` is the column name. Its format is "c{col_pos}".

        -   ``@colwidth`` width of the column (with its unit).
            The unit is defined by the *width_unit* options.

        .. note::

           The ``@colnum`` attribute (number of column) is not generated
           because this value is usually implied, and can be deduce
           from the ``@colname`` attribute.

        :type  group_elem: etree._Element
        :param group_elem: Parent element: ``<tgroup>``.

        :type  col: benker.table.ColView
        :param col: Columns
        """
        col_styles = col.styles
        attrs = {u'colname': u"c{0}".format(col.col_pos)}
        if 'width' in col_styles:
            width = col_styles['width']
            width, unit = re.findall(r"([+-]?(?:[0-9]*[.])?[0-9]+)(\w+)", width)[0]
            value = convert_value(float(width), unit, self.width_unit)
            attrs['colwidth'] = u"{value:0.2f}{unit}".format(value=value, unit=self.width_unit)
        etree.SubElement(group_elem, u"colspec", attrib=attrs)

    def build_tbody(self, group_elem, row_list, nature_tag):
        """
        Build the CALS ``<tbody>``, `<thead>``, or `<tfoot>`` element.

        :type  group_elem: etree._Element
        :param group_elem: Parent element: ``<tgroup>``.

        :param row_list: List of rows

        :param nature_tag: name of the tag: 'tbody', 'thead' or 'tfoot'.
        """
        nature_elem = etree.SubElement(group_elem, nature_tag)
        for row in row_list:
            self.build_row(nature_elem, row)

    def build_row(self, tbody_elem, row):
        """
        Build the CALS ``<row>`` element.

        CALS attributes:

        -   ``@valign`` is built from the "valign" style.
            Values can be "top", "middle", "bottom" (note: "baseline" is not supported).
            Default value is "bottom".

        .. note::

           A row can be marked as inserted if "x-ins" is defined in the row styles.
           Revision marks are inserted before and after a ``<row>``
           using a couple of processing-instructions.
           We use the ``<?change-start?>`` PI to mark the start of the inserted row,
           and the ``<?change-end?>`` PI to mark the end.

        :type  tbody_elem: etree._Element
        :param tbody_elem: Parent element: ``<tbody>``, `<thead>``, or `<tfoot>``.

        :type  row: benker.table.RowView
        :param row: The row.
        """
        # Possible attrs:
        # - Row height: min-height' or 'height'
        # - Insertion revision mark: 'x-ins', 'x-ins-id', 'x-ins-author', 'x-ins-date'
        # - Vertical align: 'valign'
        #
        row_styles = row.styles
        attrs = {}
        if 'valign' in row_styles:
            # same values as CSS/Properties/vertical-align
            attrs['valign'] = {'top': 'top',
                               'middle': 'middle',
                               'bottom': 'bottom',
                               'baseline': 'bottom'}[row_styles['valign']]

        if 'x-ins' in row_styles:
            # <?change-start change-id="ct140446841083680" type="row:insertion"
            #   creator="Anita BARREL" date="2017-11-15T11:46:00"?>
            rev_attrs = collections.OrderedDict({'type': 'row:insertion'})
            if 'x-ins-id' in row_styles:
                rev_attrs['change-id'] = "ct{0}".format(row_styles['x-ins-id'])
            if 'x-ins-author' in row_styles:
                rev_attrs['creator'] = row_styles['x-ins-author']
            if 'x-ins-date' in row_styles:
                rev_attrs['date'] = row_styles['x-ins-date']
            rev_pi = revision_mark("change-start", rev_attrs)
            tbody_elem.append(rev_pi)

        row_elem = etree.SubElement(tbody_elem, u"row", attrib=attrs)

        if 'x-ins' in row_styles:
            # <?change-end change-id="ct139821811327752" type="row:insertion"?>
            rev_attrs = collections.OrderedDict({'type': 'row:insertion'})
            if 'x-ins-id' in row_styles:
                rev_attrs['change-id'] = "ct{0}".format(row_styles['x-ins-id'])
            rev_pi = revision_mark('change-end', rev_attrs)
            tbody_elem.append(rev_pi)

        for cell in row.owned_cells:
            self.build_cell(row_elem, cell)

    # noinspection PyMethodMayBeStatic
    def build_cell(self, row_elem, cell):
        """
        Build the CALS ``<entry>`` element.

        CALS attributes:

        -   ``@colsep`` is built from the "border-right" style.
            Default value is "1" (displayed), so, it is better to always define it.
            This value is only set if different from the table ``@colsep`` value.

        -   ``@rowsep`` is built from the "border-bottom" style.
            Default value is "1" (displayed), so, it is better to always define it.
            This value is only set if different from the table ``@rowsep`` value.

        -   ``@valign`` is built from the "valign" style.
            Values can be "top", "middle", "bottom" (note: "baseline" is not supported).
            Default value is "bottom".

        -   ``@align`` is built from the "align" style.
            Values can be "left", "center", "right", or "justify".
            Default value is "left".
            note: paragraphs alignment should be preferred to cells alignment.

        -   ``@namest``/``@nameend`` are set when the cell is spanned horizontally.

        -   ``@morerows`` is set when the cell is spanned vertically.

        :type  row_elem: etree._Element
        :param row_elem: Parent element: ``<row>``.

        :type  cell: benker.cell.Cell
        :param cell: The cell.
        """
        cell_styles = cell.styles
        attrs = {}
        if cell.box.max.x != self._table.bounding_box.max.x:
            # generate @colsep if the cell isn't in the last column
            cell_colsep = _get_colsep_attr(cell_styles, "border-right")
            if cell_colsep and cell_colsep != self._table_colsep:
                attrs['colsep'] = cell_colsep
        if cell.box.max.y != self._table.bounding_box.max.y:
            # generate @rowsep if the cell isn't in the last row
            cell_rowsep = _get_rowsep_attr(cell_styles, "border-bottom")
            if cell_rowsep and cell_rowsep != self._table_rowsep:
                attrs['rowsep'] = cell_rowsep
        if 'vertical-align' in cell_styles:
            # same values as CSS/Properties/vertical-align
            # 'w-both' is an extension of OoxmlParser
            attrs['valign'] = {'top': u'top',
                               'middle': u'middle',
                               'bottom': u'bottom',
                               'baseline': u'bottom',
                               'w-both': u'bottom'}[cell_styles['vertical-align']]
        if 'align' in cell_styles:
            # same values as CSS/Properties/text-align
            attrs['align'] = {'left': u'left',
                              'center': u'center',
                              'right': u'right',
                              'justify': u'justify'}[cell_styles['align']]
        if cell.width > 1:
            attrs[u"namest"] = u"c{0}".format(cell.box.min.x)
            attrs[u"nameend"] = u"c{0}".format(cell.box.max.x)
        if cell.height > 1:
            attrs[u"morerows"] = str(cell.height - 1)

        entry_elem = etree.SubElement(row_elem, u"entry", attrib=attrs)
        if cell.content is not None:
            entry_elem.extend(cell.content)



# coding: utf-8
"""
.. _available-converters:

Available Converters
====================

This package contains a collection of converters
which you can use to convert tables from one format to another.

To do that, you need to select a **parser** and a **builder**:

- You can pick a parser in the :ref:`Available Parsers <available-parsers>`.
- You can pick a builder in the :ref:`Available Builders <available-builders>`.
"""



# coding: utf-8
"""
Grid
====

A grid of cells.

Example of grid::

.. doctest:: grid_demo

    >>> from benker.grid import Grid
    >>> from benker.cell import Cell

    >>> grid = Grid()
    >>> grid[1, 1] = Cell("red", height=2)
    >>> grid[2, 1] = Cell("pink", width=2)
    >>> grid[2, 2] = Cell("blue")

    >>> print(grid)
    +-----------+-----------------------+
    |    red    |   pink                |
    |           +-----------+-----------+
    |           |   blue    |           |
    +-----------+-----------+-----------+

You can retrieve the grid cells as follow:

.. doctest:: grid_demo

    >>> from benker.grid import Grid
    >>> from benker.cell import Cell

    >>> grid = Grid()
    >>> grid[1, 1] = Cell("red", height=2)
    >>> grid[2, 1] = Cell("pink", width=2)
    >>> grid[2, 2] = Cell("blue")

    >>> grid[1, 1]
    <Cell('red', styles={}, nature=None, x=1, y=1, width=1, height=2)>
    >>> grid[2, 1]
    <Cell('pink', styles={}, nature=None, x=2, y=1, width=2, height=1)>
    >>> grid[2, 2]
    <Cell('blue', styles={}, nature=None, x=2, y=2, width=1, height=1)>
    >>> grid[3, 3]
    Traceback (most recent call last):
        ...
    KeyError: Coord(x=3, y=3)

A grid has a bounding box, useful to get the grid sizes:

.. doctest:: grid_demo

    >>> from benker.grid import Grid
    >>> from benker.cell import Cell

    >>> grid = Grid()
    >>> grid[1, 1] = Cell("red", height=2)
    >>> grid[2, 1] = Cell("pink", width=2)
    >>> grid[2, 2] = Cell("blue")

    >>> grid.bounding_box
    Box(min=Coord(x=1, y=1), max=Coord(x=3, y=2))
    >>> grid.bounding_box.size
    Size(width=3, height=2)

You can expand the cell size horizontally or vertically:

.. doctest:: grid_demo

    >>> from benker.grid import Grid
    >>> from benker.cell import Cell

    >>> grid = Grid()
    >>> grid[1, 1] = Cell("red", height=2)
    >>> grid[2, 1] = Cell("pink", width=2)
    >>> grid[2, 2] = Cell("blue")

    >>> grid.expand((2, 2), width=1)
    <Cell('blue', styles={}, nature=None, x=2, y=2, width=2, height=1)>
    >>> print(grid)
    +-----------+-----------------------+
    |    red    |   pink                |
    |           +-----------------------+
    |           |   blue                |
    +-----------+-----------------------+

The content of the merged cells is merged too:

.. doctest:: grid_demo

    >>> from benker.grid import Grid
    >>> from benker.cell import Cell

    >>> grid = Grid()
    >>> grid[1, 1] = Cell("red", height=2)
    >>> grid[2, 1] = Cell("pink", width=2)
    >>> grid[2, 2] = Cell("blue", width=2)

    >>> grid.merge((2, 1), (3, 2), content_appender=lambda a, b: "/".join([a, b]))
    <Cell('pink/blue', styles={}, nature=None, x=2, y=1, width=2, height=2)>
    >>> print(grid)
    +-----------+-----------------------+
    |    red    | pink/blue             |
    |           |                       |
    |           |                       |
    +-----------+-----------------------+
"""
import bisect
import itertools
import operator

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

from benker.box import Box
from benker.coord import Coord
from benker.drawing import draw
from benker.size import Size


class Grid(MutableMapping):
    """
    Collection of :class:`~benker.cell.Cell` objects ordered in a grid of rows and columns.
    """
    __slots__ = ('_cells',)

    def __init__(self, cells=None):
        """
        Construct a grid from an iterable of cells.
        All cells must be disjoints.

        Cells are ordered according to the total ordering of the cell boxes.

        :type  cells: typing.Iterable[benker.cell.Cell]
        :param cells: Collection of cells.

        :raises KeyError: if at least one cell intersect another one.
        """
        self._cells = []
        cells = cells or []
        for cell in cells:
            self.__setitem__(cell.min, cell)

    def __repr__(self):
        cls = self.__class__.__name__
        return "{cls}({cells!r})".format(cls=cls, cells=self._cells)

    def __contains__(self, coord):
        coord = Coord.from_value(coord)
        return any(coord in cell for cell in self._cells)

    def __getitem__(self, coord):
        coord = Coord.from_value(coord)
        for cell in self._cells:
            if coord in cell:
                return cell
        raise KeyError(coord)

    def __delitem__(self, coord):
        coord = Coord.from_value(coord)
        for cell in self._cells[:]:
            if coord in cell:
                self._cells.remove(cell)
                return
        raise KeyError(coord)

    def __setitem__(self, coord, new_cell):
        coord = Coord.from_value(coord)  # type: Coord
        new_cell = new_cell.move_to(coord)
        for cell in self._cells[:]:
            if cell.box.intersect(new_cell.box):
                raise KeyError(coord)
        boxes = [cell.box for cell in self._cells]
        index = bisect.bisect_left(boxes, Box(new_cell.box.min))
        self._cells.insert(index, new_cell)

    def __len__(self):
        return len(self._cells)

    def __iter__(self):
        return iter(self._cells)

    def __str__(self):
        return draw(self)

    @property
    def bounding_box(self):
        """ Bounding box of the grid (``None`` if the grid is empty). """
        if len(self):
            boxes = [cell.box for cell in self._cells]
            bounding_box = boxes[0].union(*boxes[1:])
            return bounding_box
        return None

    def merge(self, start, end, content_appender=None):
        """
        Merge a group of cells contained in a bounding box,
        using the *content_appender* to append cell contents.

        The coordinates *start* and *end* delimit a group of cells to merge.

        .. warning::

           All the cells of the group must be included in the group bounding box,
           no intersection is allowed. If not, :class:`ValueError` is raised.

        See also the method :meth:`~benker.grid.Grid.expand`
        to expand (or shrink) the width and/or height of a cell.

        :type  start: Coord or tuple[int, int]
        :param start:
            Top-left coordinates of the group of cells to merge.

        :type  start: Coord or tuple[int, int]
        :param end:
            Bottom-right coordinates of the group of cells to merge (inclusive).

        :param content_appender:
            Function to use to append the cell contents.
            The function must have the following signature: ``f(a, b) -> c``,
            where *a*, *b* anc *c* must be of the same type than the cell content.
            If not provided, the default function is :func:`operator.__add__`.

        :return: The merged cell.

        :raises ValueError:
            If the group of cells is empty or if cells cannot be merged.
        """
        start_coord = Coord.from_value(start)
        end_coord = Coord.from_value(end)
        content_appender = content_appender or operator.__add__
        new_box = Box(start_coord, end_coord)
        merged_cells = []
        unchanged_cells = []
        for cell in self._cells:
            if cell.box in new_box:
                merged_cells.append(cell)
            elif cell.box.intersect(new_box):
                raise ValueError((start, end))
            else:
                unchanged_cells.append(cell)
        if not merged_cells:
            # nothing to merge
            raise ValueError((start, end))
        first = merged_cells.pop(0)
        new_cell = first.transform(coord=new_box.min, size=new_box.size)
        for cell in merged_cells:
            if new_cell.content is None:
                new_cell.content = cell.content
            elif cell.content is None:
                pass  # no change
            else:
                new_cell.content = content_appender(new_cell.content, cell.content)
            new_cell.styles.update(cell.styles)
        self._cells = unchanged_cells
        boxes = [cell.box for cell in self._cells]
        index = bisect.bisect_left(boxes, Box(new_cell.box.min))
        self._cells.insert(index, new_cell)
        return new_cell

    def expand(self, coord, width=0, height=0, content_appender=None):
        """
        Expand (or shrink) the *width* and/or *height* of a cell,
        using the *content_appender* to append cell contents.

        See also the method :meth:`~benker.grid.Grid.merge`
        to merge a group of cells contained in a bounding box.

        :param coord:
            Coordinates of the cell to expand (or shrink).

        :param width:
            Number of columns to add to the cell width.

        :param height:
            Number of rows to add to the cell height.

        :param content_appender:
            Function to use to append the cell contents.
            The function must have the following signature: ``f(a, b) -> c``,
            where *a*, *b* anc *c* must be of the same type than the cell content.
            If not provided, the default function is :func:`operator.__add__`.

        :return: The merged cell.

        :raises ValueError:
            If the group of cells is empty or if cells cannot be merged.
        """
        content_appender = content_appender or operator.__add__
        box = self[coord].box
        start = box.min
        end = box.max + Size(width, height)
        return self.merge(start, end, content_appender=content_appender)

    def iter_rows(self):
        """ Iterate the cells grouped by rows. """
        cells = self._cells
        for group, cells in itertools.groupby(cells, key=lambda c: c.min.y):
            yield tuple(cells)



# coding: utf-8
"""
Formex4 Builder
===============

This module can construct a Formex4 table from
an instance of type :class:`~benker.table.Table`.

:abbr:`Formex (Formalized Exchange of Electronic Publications)` describes the format for the exchange
of data between the Publication Office and its contractors. In particular, it defines the logical
markup for documents which are published in the different series of the Official Journal
of the European Union.

This builder allow you to convert Word document tables
into Formex4 tables using the Formex4 schema (formex-05.59-20170418.xd).

Specifications and examples:

- The Formex4 documentation and schema is available online in the Publication Office:
  `Formex Version 4 <https://publications.europa.eu/documents/3938058/5910419/formex_manual_on_screen_version.html>`_.

- An example of Formex4 table is available in the Schema documentation:
  `TBL <https://publications.europa.eu/documents/3938058/5910419/formex_manual_on_screen_version.html#TBL>`_

"""
import collections

from lxml import etree

from benker.builders.base_builder import BaseBuilder
from benker.parsers.lxml_iterwalk import iterwalk

text_type = type(u"")


def revision_mark(name, attrs):
    target = etree.tounicode(etree.Element(name, attrib=attrs))
    target = target.replace("<{0} ".format(name), "").replace("/>", "")
    rev_pi = etree.ProcessingInstruction(name, target)
    return rev_pi


class Formex4Builder(BaseBuilder):
    """
    Formex4 builder used to convert tables into ``TBL`` elements according to the
    `TBL Schema <https://publications.europa.eu/documents/3938058/5910419/formex_manual_on_screen_version.html#TBL>`_
    """

    def __init__(self, detect_titles=False, **options):
        """
        Initialize the builder.

        :param detect_titles:
            If this option is enable, a title will be generated if the first row
            contains an unique cell with centered text.

        :keyword options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.

        .. versionchanged:: 0.4.3
           Add the *detect_titles* option.
        """
        # Internal state of the table used during building
        self._table = None
        self._no_seq = 0
        # options
        self.detect_titles = detect_titles
        super(Formex4Builder, self).__init__(**options)

    def generate_table_tree(self, table):
        """
        Build the XML table from the Table instance.

        :type  table: benker.table.Table
        :param table: Table

        :return: Table tree
        """
        table_elem = self.build_tbl(table)
        return table_elem

    def build_tbl(self, table):
        """
        Build the Formex4 ``<TBL>`` element.

        Formex4 attributes:

        -   ``@NO.SEQ`` This mandatory attribute provides a sequence number to the table.
            This number represents the order in which the table appears in the document.

        -   ``@CLASS`` The CLASS attribute is mandatory and is used to specify the type of data
            contained in the table. The allowed values are:

            *   GEN: if the table contains general data (default value),
            *   SCHEDULE: if it is a schedule,
            *   RECAP: if it is a synoptic table.

            These two last values are only used for documents related to the general budget.

        -   ``@COLS`` This mandatory attribute provides the actual number of columns of the table.

        -   ``@PAGE.SIZE`` The PAGE.SIZE attribute takes one of these values:

            *   DOUBLE.LANDSCAPE: table on two A4 pages forming an A3 landscape page,
            *   DOUBLE.PORTRAIT: table on two A4 pages forming an A3 portrait page,
            *   SINGLE.LANDSCAPE: table on a single A4 page in landscape,
            *   SINGLE.PORTRAIT: table on a single A4 page in portrait (default).

        :type  table: benker.table.Table
        :param table: Table

        :return: The newly-created ``<TBL>`` element.
        """
        self._table = table
        table_styles = table.styles
        self._no_seq += 1
        attrs = {
            'NO.SEQ': u"{:04d}".format(self._no_seq),
            # 'CLASS': u"GEN",  # default value
            'COLS': str(len(table.cols)),
            # 'PAGE.SIZE': u"SINGLE.PORTRAIT",  # default value
        }
        if 'x-sect-orient' in table_styles:
            # size = (width x height) in 'pt'
            size = table_styles.get('x-sect-size', (595, 842))
            # C4 format 22.9cm x 32.4cm is a little bigger than A4
            if (size[0] <= 649 and size[1] <= 918) or (size[0] <= 918 and size[1] <= 649):
                orient_page_sizes = {"landscape": "SINGLE.LANDSCAPE"}
            else:
                orient_page_sizes = {"landscape": "DOUBLE.LANDSCAPE", "portrait": "DOUBLE.PORTRAIT"}
            orient = table_styles['x-sect-orient']
            if orient in orient_page_sizes:
                attrs['PAGE.SIZE'] = orient_page_sizes[orient]
        table_elem = etree.Element(u"TBL", attrib=attrs)
        self.build_corpus(table_elem, table)
        return table_elem

    def build_corpus(self, tbl_elem, table):
        """
        Build the Formex4 ``<CORPUS>`` element.

        :type  tbl_elem: etree._Element
        :param tbl_elem: Parent element: ``<TBL>``.

        :type  table: benker.table.Table
        :param table: Table

        .. versionchanged:: 0.4.3
            If this option *detect_titles* is enable, a title will be generated
            if the first row contains an unique cell with centered text.
        """
        # table_styles = table.styles
        attrs = {}  # no attribute
        rows = list(table.rows)

        if self.detect_titles:
            # Does the first row/cell contains a centered title?
            first_cell = table[(1, 1)]
            align = first_cell.styles.get('align')
            if first_cell.width == table.bounding_box.width and align == "center":
                # yes, we can generate the title
                self.build_title(tbl_elem, rows.pop(0))

        corpus_elem = etree.SubElement(tbl_elem, u"CORPUS", attrib=attrs)
        for row in rows:
            self.build_row(corpus_elem, row)

    def build_title(self, tbl_elem, row):
        """
        Build the table title using the ``<TITLE>`` element.

        For instance:

        .. code-block:: xml

           <TITLE>
             <TI>
               <P>Table IV</P>
             </TI>
           </TITLE>

        :type  tbl_elem: etree._Element
        :param tbl_elem: Parent element: ``<TBL>``.

        :type  row: benker.table.RowView
        :param row: The row which contains the title.

        .. versionchanged:: 0.4.4
           Modification of the Formex4 builder to better deal with empty cells (management of ``<IE/>`` tags).
        """
        title_elem = etree.SubElement(tbl_elem, u"TITLE")
        for cell in row.owned_cells:
            # When a cell is empty, we need to insert the ``<IE/>`` tag.
            is_empty_cell = cell.styles.get("x-cell-empty", "false") == "true"
            # We can also have an empty content if a short row is completed by empty cells.
            if is_empty_cell or cell.content is None or cell.content == "":
                # assert cell.content in {None, "", []}
                ti_elem = etree.SubElement(title_elem, u"TI")
                etree.SubElement(ti_elem, u"IE")
            else:
                if isinstance(cell.content, type(u"")):
                    # mainly useful for unit test
                    ti_elem = etree.SubElement(title_elem, u"TI")
                    p_elem = etree.SubElement(ti_elem, u"P")
                    p_elem.text = cell.content
                else:
                    paragraphs = list(cell.content)
                    ti_elem = etree.SubElement(title_elem, u"TI")
                    ti_elem.append(paragraphs[0])
                    sti_elem = etree.SubElement(title_elem, u"STI")
                    sti_elem.extend(paragraphs[1:])

    def build_row(self, corpus_elem, row):
        """
        Build the Formex4 ``<ROW>`` element.

        Formex4 attributes:

        -   ``@TYPE`` The TYPE attribute indicates the specific role of the row in the table.
            The allowed values are:

            *   ALIAS: if the row contains aliases. Such references may be used when the table
                is included on several pages of a publication. The references are associated
                to column headers on the first page and are repeated on subsequent pages.
            *   HEADER: if the row contains cells which may be considered as a column header.
                This generally occurs for the first row of a table.
            *   NORMAL: if most of the cells of the row contain 'simple' or 'normal' data.
                This is the default value.
            *   NOTCOL: if the cells of the row contain units of measure relating to subsequent rows.
            *   TOTAL: if the row contains data which could be considered as 'totals'.

            Note that this TYPE attribute is also provided for the cells (CELL), which could
            be used to override the value defined for the row. On the other hand, 'NORMAL'
            is the default value, so it is necessary to specify the TYPE attribute value
            in each cell of a row which has a specific type in order to avoid the default
            overriding (see the first row of the example below).

        :type  corpus_elem: etree._Element
        :param corpus_elem: Parent element: ``<CORPUS>``.

        :type  row: benker.table.RowView
        :param row: The row.
        """
        row_styles = row.styles
        attrs = {}
        nature_types = {"head": u"HEADER", "foot": u"TOTAL"}
        if row.nature in nature_types:
            attrs['TYPE'] = nature_types[row.nature]

        if 'x-ins' in row_styles:
            # <?change-start change-id="ct140446841083680" type="row:insertion"
            #   creator="Anita BARREL" date="2017-11-15T11:46:00"?>
            rev_attrs = collections.OrderedDict({'type': 'row:insertion'})
            if 'x-ins-id' in row_styles:
                rev_attrs['change-id'] = "ct{0}".format(row_styles['x-ins-id'])
            if 'x-ins-author' in row_styles:
                rev_attrs['creator'] = row_styles['x-ins-author']
            if 'x-ins-date' in row_styles:
                rev_attrs['date'] = row_styles['x-ins-date']
            rev_pi = revision_mark("change-start", rev_attrs)
            corpus_elem.append(rev_pi)

        row_elem = etree.SubElement(corpus_elem, u"ROW", attrib=attrs)

        if 'x-ins' in row_styles:
            # <?change-end change-id="ct139821811327752" type="row:insertion"?>
            rev_attrs = collections.OrderedDict({'type': 'row:insertion'})
            if 'x-ins-id' in row_styles:
                rev_attrs['change-id'] = "ct{0}".format(row_styles['x-ins-id'])
            rev_pi = revision_mark('change-end', rev_attrs)
            corpus_elem.append(rev_pi)

        for cell in row.owned_cells:
            self.build_cell(row_elem, cell, row)

    # noinspection PyMethodMayBeStatic
    def build_cell(self, row_elem, cell, row):
        """
        Build the Formex4 ``<CELL>`` element.

        Formex4 attributes:

        -   ``@COL`` The mandatory COL attribute is used to specify in which column the cell is located.

        -   ``@COLSPAN`` When a cell in a row 'A' must be linked to a group of cells in the same row,
            the first CELL element of this group has to provide the COLSPAN attribute. The value of
            the COLSPAN attribute is the number of cells in the group. The COL attribute of the first
            cell indicates the number of the first column in the group.

            The use of the COLSPAN attribute is only allowed to relate the value of a cell in several
            columns within the same row. Its value must be at least equal to '2'.

        -   ``@ROWSPAN`` When a cell in column 'A' is linked to a cell in row 'B' located just below
            row 'A', the CELL element of this single cell must provide the ROWSPAN attribute.
            The value of the ROWSPAN attribute is equal to the number of cells in the group.
            The CELL element relating to the single cell must be placed within the first ROW element
            in the group. The ROW elements corresponding to the other rows in the group may not
            contain any CELL elements for the column containing the single cell 'A'.

            The use of the ROWSPAN attribute is only authorised to relate the value of a cell in several
            rows. Its value must be at least equal to '2'.

        -   ``@ACCH`` If the group of related cells is physically delimited by a horizontal brace,
            this symbol must be marked up using the ACCH attribute.

        -   ``@ACCV`` If the group of related cells is physically delimited by a vertical brace,
            this symbol must be marked up using the ACCV attribute.

        -   ``@TYPE`` The TYPE attribute of the CELL element is used to indicate locally the type
            of contents of the cells. It overrides the value of the TYPE attribute defined for
            the row (ROW) which contains the cell.

        :type  row_elem: etree._Element
        :param row_elem: Parent element: ``<ROW>``.

        :type  cell: benker.cell.Cell
        :param cell: The cell.

        :type  row: benker.table.RowView
        :param row: The parent row.

        .. versionchanged:: 0.4.4
           Modification of the Formex4 builder to better deal with empty cells (management of ``<IE/>`` tags).
        """
        # cell_styles = cell.styles
        attrs = {'COL': str(cell.box.min.x)}
        if cell.nature and cell.nature != row.nature:
            nature_types = {"head": u"HEADER", "body": u"NORMAL", "foot": u"TOTAL"}
            attrs['TYPE'] = nature_types[cell.nature]
        if cell.width > 1:
            attrs[u"COLSPAN"] = str(cell.width)
        if cell.height > 1:
            attrs[u"ROWSPAN"] = str(cell.height)
        cell_elem = etree.SubElement(row_elem, u"CELL", attrib=attrs)

        # When a cell is empty, we need to insert the ``<IE/>`` tag.
        is_empty_cell = cell.styles.get("x-cell-empty", "false") == "true"
        # We can also have an empty content if a short row is completed by empty cells.
        if is_empty_cell or cell.content is None or cell.content == "":
            # The IE element is used to explicitly indicate
            # that specific structures have an empty content.
            etree.SubElement(cell_elem, u"IE")
        else:
            if isinstance(cell.content, type(u"")):
                # mainly useful for unit test
                cell_elem.text = cell.content
            else:
                cell_elem.extend(cell.content)

    def finalize_tree(self, tree):
        """
        Finalize the resulting tree structure:
        calculate the ``@NO.SEQ`` values: sequence number of each table.

        :type  tree: etree._ElementTree
        :param tree: The resulting tree.
        """
        root = tree.getroot()
        context = iterwalk(root, events=('start',), tag=('TBL',))

        stack = []
        for action, elem in context:  # type: str, etree._Element
            elem_tag = elem.tag
            if elem_tag == 'TBL':
                elem_level = int(elem.xpath("count(ancestor-or-self::TBL)"))
                curr_level = len(stack)
                if curr_level < elem_level:
                    stack.extend([0] * (elem_level - curr_level))
                else:
                    stack[:] = stack[:elem_level]
                stack[elem_level - 1] += 1
                no_seq = u'.'.join(u"{:04d}".format(value) for value in stack)
                elem.attrib['NO.SEQ'] = no_seq
            else:
                raise NotImplementedError(elem_tag)



# coding: utf-8
"""
Cell Coordinates
================

A :class:`~benker.coord.Coord` object is used to represent the *x* and *y* positions of a :class:`~benker.cell.Cell`.
The *x* is the left position (column number) and the *y* is the top position (row number).
The default cell coordinates is (1, 1).

This module defines the :class:`~benker.coord.Coord` tuple and give some classic use cases.
"""
import collections

from benker.alphabet import int_to_alphabet
from benker.size import Size

CoordTuple = collections.namedtuple('CoordTuple', ['x', 'y'])


class Coord(CoordTuple):
    """
    Coordinates of a cell in a grid: *x* is the left column, *y* if the top row.

    Usage:

    .. doctest:: coord_demo

        >>> from benker.coord import Coord

        >>> coord = Coord(5, 3)
        >>> coord
        Coord(x=5, y=3)
        >>> coord.x
        5
        >>> coord.y
        3
        >>> str(coord)
        'E3'

    You can use the "+" or "-" operators to move the coordinates:

    .. doctest:: coord_demo

        >>> from benker.size import Size

        >>> Coord(2, 1) + Size(3, 3)
        Coord(x=5, y=4)
        >>> Coord(5, 4) - Size(3, 3)
        Coord(x=2, y=1)

    .. warning::

        You cannot add or subtract two coordinates, only a coordinate and a size.

        .. doctest:: coord_demo

            >>> from benker.coord import Coord

            >>> Coord(2, 1) + Coord(3, 3)
            Traceback (most recent call last):
                ...
            TypeError: <class 'benker.coord.Coord'>
    """
    __slots__ = ()

    def __str__(self):
        return int_to_alphabet(self.x) + str(self.y)

    def __repr__(self):
        return super(Coord, self).__repr__().replace('CoordTuple', 'Coord')

    def __add__(self, size):
        size_type = type(size)
        if size_type is Size:
            return Coord(self.x + size.width, self.y + size.height)
        elif size_type is tuple:
            return Coord(self.x + size[0], self.y + size[1])
        elif size_type is int:
            return Coord(self.x + size, self.y + size)
        else:
            raise TypeError(repr(size_type))

    def __sub__(self, size):
        size_type = type(size)
        if size_type is Size:
            return Coord(self.x - size.width, self.y - size.height)
        elif size_type is tuple:
            return Coord(self.x - size[0], self.y - size[1])
        elif size_type is int:
            return Coord(self.x - size, self.y - size)
        else:
            raise TypeError(repr(size_type))

    @classmethod
    def from_value(cls, value):
        """
        Convert a value of type :class:`tuple` to a :class:`~benker.coord.Coord` tuple.

        :param value: tuple of two integers or :class:`~benker.coord.Coord` tuple.

        :return: Newly created object.

        :raises TypeError:
            if the value is not a tuple of integers nor a :class:`~benker.coord.Coord` tuple.
        """
        value_type = type(value)
        if value_type is cls:
            return value
        elif value_type is tuple and tuple(map(type, value)) == (int, int):
            return cls(*value)
        raise TypeError(repr(value_type))



# coding: utf-8
import pytest
from lxml import etree

from benker.builders.cals import CalsBuilder
from benker.cell import Cell
from benker.table import Table

TEST_DATA__WITH_SEP = [

    pytest.param(
        {},
        {},
        id="no_style"),

    pytest.param(
        {'border-top': 'double 1.0pt #FFFFFF',
         'border-right': 'dotted 2.0pt #FF0000',
         'border-bottom': 'dashed 3.0pt #00FF00',
         'border-left': 'outset 4.0pt #0000FF'},
        {'colsep': '1', 'rowsep': '1'},
        id="style_border-all"),

    pytest.param(
        {'border-top': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-top"),

    pytest.param(
        {'border-right': 'double 1.0pt #FFFFFF'},
        {'colsep': '1'},
        id="style_border-right"),

    pytest.param(
        {'border-bottom': 'double 1.0pt #FFFFFF'},
        {'rowsep': '1'},
        id="style_border-bottom"),

    pytest.param(
        {'border-left': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-left"),
]


@pytest.mark.parametrize('cell_styles, expected', TEST_DATA__WITH_SEP)
def test_build_cell__with_sep(cell_styles, expected):
    # .. note::
    #
    #   CALS default value for @colsep and @rowsep is "1".

    builder = CalsBuilder()

    # -- create a minimal <row> element
    row_elem = etree.XML("<row/>")

    # -- setup a minimal cell
    p_elem = etree.XML("<p>text</p>")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, styles=cell_styles)
    cell_x2_y1 = Cell([p_elem], x=2, y=1, styles=cell_styles)
    cell_x1_y2 = Cell([p_elem], x=1, y=2, styles=cell_styles)
    cell_x2_y2 = Cell([p_elem], x=2, y=2, styles=cell_styles)
    builder._table = Table([cell_x1_y1, cell_x2_y1, cell_x1_y2, cell_x2_y2])

    # -- build the cell
    builder.build_cell(row_elem, cell_x1_y1)

    # -- check the '<entry>' attributes
    entry_elem = row_elem[0]  # type: etree._Element
    assert entry_elem.tag == u"entry"
    assert entry_elem.attrib == expected
    assert entry_elem[0] == p_elem


TEST_DATA__WITHOUT_SEP = [

    pytest.param(
        {},
        {},
        id="no_style"),

    pytest.param(
        {'border-top': 'double 1.0pt #FFFFFF',
         'border-right': 'dotted 2.0pt #FF0000',
         'border-bottom': 'dashed 3.0pt #00FF00',
         'border-left': 'outset 4.0pt #0000FF'},
        {'colsep': '1', 'rowsep': '1'},
        id="style_border-all"),

    pytest.param(
        {'border-top': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-top"),

    pytest.param(
        {'border-right': 'double 1.0pt #FFFFFF'},
        {'colsep': '1'},
        id="style_border-right"),

    pytest.param(
        {'border-bottom': 'double 1.0pt #FFFFFF'},
        {'rowsep': '1'},
        id="style_border-bottom"),

    pytest.param(
        {'border-left': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-left"),
]


@pytest.mark.parametrize('cell_styles, expected', TEST_DATA__WITHOUT_SEP)
def test_build_cell__without_sep(cell_styles, expected):
    # .. note::
    #
    #   CALS default value for @colsep and @rowsep is "1"
    #   For testing, we set it to "0" by default.

    builder = CalsBuilder()
    builder._table_colsep = u"0"
    builder._table_rowsep = u"0"

    # -- create a minimal <row> element
    row_elem = etree.XML("<row/>")

    # -- setup a minimal cell
    p_elem = etree.XML("<p>text</p>")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, styles=cell_styles)
    cell_x2_y1 = Cell([p_elem], x=2, y=1, styles=cell_styles)
    cell_x1_y2 = Cell([p_elem], x=1, y=2, styles=cell_styles)
    cell_x2_y2 = Cell([p_elem], x=2, y=2, styles=cell_styles)
    builder._table = Table([cell_x1_y1, cell_x2_y1, cell_x1_y2, cell_x2_y2])

    # -- build the cell
    builder.build_cell(row_elem, cell_x1_y1)

    # -- check the '<entry>' attributes
    entry_elem = row_elem[0]  # type: etree._Element
    assert entry_elem.tag == u"entry"
    assert entry_elem.attrib == expected
    assert entry_elem[0] == p_elem


TEST_DATA__ALIGN = [

    pytest.param(
        {'vertical-align': 'top', 'align': 'left'},
        {'valign': 'top', 'align': 'left'},
        id="style_vertical-align_top_align_left"),

    pytest.param(
        {'vertical-align': 'bottom', 'align': 'right'},
        {'valign': 'bottom', 'align': 'right'},
        id="style_vertical-align_bottom_align_right"),

    pytest.param(
        {'vertical-align': 'middle', 'align': 'center'},
        {'valign': 'middle', 'align': 'center'},
        id="style_vertical-align_middle_align_center"),

    pytest.param(
        {'vertical-align': 'w-both', 'align': 'justify'},
        {'valign': 'bottom', 'align': 'justify'},
        id="style_vertical-align_both_align_justify"),
]


@pytest.mark.parametrize('cell_styles, expected', TEST_DATA__ALIGN)
def test_build_cell__align(cell_styles, expected):
    builder = CalsBuilder()

    # -- create a minimal <row> element
    row_elem = etree.XML("<row/>")

    # -- setup a minimal cell
    p_elem = etree.XML("<p>text</p>")
    cell = Cell([p_elem], x=1, y=1, styles=cell_styles)
    builder._table = Table([cell])

    # -- build the cell
    builder.build_cell(row_elem, cell)

    # -- check the '<entry>' attributes
    entry_elem = row_elem[0]  # type: etree._Element
    assert entry_elem.tag == u"entry"

    # we don't want to check @colsep/@rowsep here
    actual = dict(entry_elem.attrib)
    actual.pop("colsep", None)
    actual.pop("rowsep", None)

    assert actual == expected
    assert entry_elem[0] == p_elem



# coding: utf-8
"""
Units
=====

Utility functions to convert values from one unit to another.
"""

#: Usual units Lengths in meter
UNITS = {
    'mm': 0.001, 'dm': 0.1, 'cm': 0.01, 'm': 1.0,
    'in': 0.001 * 25.4,
    'ft': 0.3048,
    'px': 0.001 * 25.4 / 72,
    'pt': 0.001 * 25.4 / 72,
    'pc': 0.001 * 25.4 / 12,
}


def convert_value(value, unit_in, unit_out):
    """
    Convert a value from one unit to another.

    To convert '1pt' to 'mm', you can do:

    .. doctest:: units_demo

        >>> from benker.units import convert_value

        >>> convert_value(1, 'pt', 'mm')
        0.353

    :type  value: int or float
    :param value: Value to convert

    :param unit_in: Current unit of the value.

    :param unit_out: Expected unit of the value.

    :rtype:  float
    :return: The converted value
    """
    return round(value * UNITS[unit_in] / UNITS[unit_out], 3)



# coding: utf-8
import zipfile

import py.path  # type hints
import pytest
import xmldiff.main
from lxml import etree

from benker.converters.ooxml2formex4 import convert_ooxml2formex4

from tests.resources import RESOURCES_DIR


def _check_expected(dst_xml, expected_name):
    # - Compare with expected
    xml_parser = etree.XMLParser(remove_blank_text=True)
    expected_xml = RESOURCES_DIR.join(expected_name)  # type: py.path.local
    expected_tree = etree.parse(str(expected_xml), parser=xml_parser)
    expected_elements = expected_tree.xpath("//TBL")
    dst_tree = etree.parse(str(dst_xml), parser=xml_parser)
    dst_elements = dst_tree.xpath("//TBL")
    assert len(expected_elements) == len(dst_elements)
    for dst_elem, expected_elem in zip(dst_elements, expected_elements):
        diff_list = xmldiff.main.diff_trees(dst_elem, expected_elem)
        assert not diff_list


def test_convert_ooxml2formex4__demo(tmpdir):
    # type: (py.path.local) -> None

    # - Unzip the ``.docx``
    src_zip = RESOURCES_DIR.join("ooxml/demo.docx")  # type: py.path.local
    with zipfile.ZipFile(str(src_zip)) as zf:
        zf.extractall(str(tmpdir))

    # - Source and destination paths
    src_xml = tmpdir.join("word/document.xml")  # type: py.path.local
    styles_xml = tmpdir.join("word/styles.xml")  # type: py.path.local
    dst_xml = tmpdir.join(src_zip.basename).new(ext=".formex4.xml")  # type: py.path.local

    # - Create some options and convert tables
    options = {"encoding": "utf-8", "styles_path": str(styles_xml)}
    convert_ooxml2formex4(str(src_xml), str(dst_xml), **options)
    expected_name = "ooxml2formex4/demo.xml"
    _check_expected(dst_xml, expected_name)


@pytest.mark.parametrize(
    "input_name, expected_name",
    [
        ("ooxml/misc_tables.xml", "ooxml2formex4/misc_tables.xml"),
        ("ooxml/simple_merge.xml", "ooxml2formex4/simple_merge.xml"),
        ("ooxml/table_in_table.xml", "ooxml2formex4/table_in_table.xml"),
        ("ooxml/Revision marks.xml", "ooxml2formex4/Revision marks.xml"),
        ("ooxml/empty_cells.xml", "ooxml2formex4/empty_cells.xml"),
        ("ooxml/demo_images.xml", "ooxml2formex4/demo_images.xml"),
    ],
)
def test_convert_ooxml2formex4(input_name, expected_name, tmpdir):
    # type: (str, str, py.path.local) -> None
    src_xml = RESOURCES_DIR.join(input_name)  # type: py.path.local
    dst_xml = tmpdir.join(src_xml.basename)
    convert_ooxml2formex4(str(src_xml), str(dst_xml), width_unit="pt")
    _check_expected(dst_xml, expected_name)


@pytest.mark.parametrize(
    "input_name, expected_name", [("ooxml/empty_cells.xml", "ooxml2formex4_detect_titles/empty_cells.xml")]
)
def test_convert_ooxml2formex4__detect_titles(input_name, expected_name, tmpdir):
    # type: (str, str, py.path.local) -> None
    src_xml = RESOURCES_DIR.join(input_name)  # type: py.path.local
    dst_xml = tmpdir.join(src_xml.basename)
    convert_ooxml2formex4(str(src_xml), str(dst_xml), width_unit="pt", detect_titles=True)
    _check_expected(dst_xml, expected_name)



# coding: utf-8
"""
Styled object
=============

A :class:`~benker.styled.Styled` object contains a dictionary of styles.

It is mainly used for :class:`~benker.table.Table`, :class:`~benker.table.RowView`,
:class:`~benker.table.ColView`, and :class:`~benker.cell.Cell`.

"""
import pprint


class Styled(object):
    """
    Styled object, like Table, Row, Column, or Cell objects.

    A styled object stores user-defined styles: a dictionary of key-value pairs.
    This values are useful to store some HTML-like styles (border-style,
    border-width, border-color, vertical-align, text-align, etc.).
    Of course, we are not tied to the HTML-like styles, you can use your
    own list of styles.

    .. note::

        The style dictionary is always copied: in other words, key-value pairs
        are copied but a shallow copy is done for the values (in general, it
        is not a problem if you use non-mutable values like :class:`str`).

    A styled object stores a nature: a way to distinguish the body cells,
    from the header and the footer. The default value is ``None``, but you can
    use "body", "header", "footer" or whatever is suitable for your needs.
    This kind of information is in general not stored in the styles,
    even if it is similar.

    Tables can also have a *nature*, similar to HTML ``@class`` attribute,
    you can use it do identify the styles to apply to your table.

    .. note::

        In a :class:`~benker.grid.Grid`, the :ref:`merging <benker__grid__merging>`
        of two natures is done by keeping the first nature and
        dropping the second one. In other words, the resulting nature is
        the group of the most top-left nature of the merged cells.

    """
    __slots__ = ('_styles', 'nature')

    def __init__(self, styles, nature):
        """
        Construct a styled object from a dictionary of styles.

        :type  styles: typing.Dict[str, str]
        :param styles:
            Dictionary of key-value pairs, where *keys* are the style names.

        :type nature: str
        :ivar nature:
            Cell *nature* used to distinguish the body cells, from the header and the footer.

            Table *nature* used to store a value similar to HTML ``@class`` attribute.
        """
        #: Dictionary of key-value pairs, where *keys* are the style names.
        self.styles = styles

        #: Cell *nature* used to distinguish the body cells, from the header and the footer.
        self.nature = nature

    def __str__(self):
        return str(self._styles)

    def __repr__(self):
        cls = self.__class__.__name__
        items = pprint.pformat(self._styles)
        nature = self.nature
        return "<{cls}({items}, {nature!r})>".format(cls=cls, items=items, nature=nature)

    @property
    def styles(self):
        """ Dictionary of styles: key-value pairs. """
        return self._styles

    @styles.setter
    def styles(self, styles):
        """ Setup the dictionary of styles (shallow copy of the items). """
        # each cell owns it's own copy of the styles
        self._styles = {} if styles is None else styles.copy()



# coding: utf-8
from __future__ import print_function

import sys

from lxml import etree

# noinspection PyProtectedMember
ElementTreeType = etree._ElementTree

# noinspection PyProtectedMember
ElementType = etree._Element


class CalsComparator(object):
    def compare_files(self, src_path, exp_file):
        src_tree = etree.parse(src_path)  # type: ElementTreeType
        exp_tree = etree.parse(exp_file)
        # The number of <table> must be the same
        src_tables = src_tree.xpath("//table")
        exp_tables = exp_tree.xpath("//table")
        self._compare_children(src_tables, exp_tables)

    def _compare_children(self, src_children, exp_children):
        assert len(src_children) == len(exp_children)
        for src_table, exp_table in zip(src_children, exp_children):
            method_name = "compare_{tag}".format(tag=src_table.tag)
            if hasattr(self, method_name):
                compare_method = getattr(self, method_name)
                compare_method(src_table, exp_table)

    def _compare_elements(self, src_tree, exp_tree):
        # type: (ElementType, ElementType) -> None
        try:
            assert src_tree.tag == exp_tree.tag
            src_keys = sorted(src_tree.attrib.keys())
            exp_keys = sorted(exp_tree.attrib.keys())
            assert src_keys == exp_keys, "src: {0!r}, exp: {1!r}".format(src_keys, exp_keys)
            src_attrs = sorted(src_tree.attrib.items())
            exp_attrs = sorted(exp_tree.attrib.items())
            assert src_attrs == exp_attrs, "src: {0!r}, exp: {1!r}".format(src_attrs, exp_attrs)
        except AssertionError:
            tree = exp_tree.getroottree()
            src_proto = etree.Element(src_tree.tag, attrib=src_tree.attrib)
            exp_proto = etree.Element(exp_tree.tag, attrib=exp_tree.attrib)
            print("FAIL: invalid xpath: {0}".format(tree.getpath(exp_tree)), file=sys.stderr)
            print("source:   " + etree.tounicode(src_proto, with_tail=False), file=sys.stderr)
            print("expected: " + etree.tounicode(exp_proto, with_tail=False), file=sys.stderr)
            raise
        else:
            src_children = list(src_tree.iterchildren('*'))
            exp_children = list(exp_tree.iterchildren('*'))
            self._compare_children(exp_children, src_children)

    compare_table = _compare_elements
    compare_tgroup = _compare_elements
    compare_colspec = _compare_elements
    compare_tbody = _compare_elements
    compare_thead = _compare_elements
    compare_tfoot = _compare_elements
    compare_row = _compare_elements
    compare_entry = _compare_elements



"""Organize your data in a spatial grid system for CALS, HTML, Formex4, Office Open XML tables conversion"""

__version__ = "0.4.4"
__author__ = "Laurent LAPORTE <laurent.laporte.pro@gmail.com>"
__date__ = "2021-11-10"
__credit__ = "(c) 2018-2021 Laurent LAPORTE"



# coding: utf-8
u"""
Box
===

A *Box* is a rectangular area defined by two coordinates:

- the top-left corner of the rectangle: the *min* coord,
- the bottom-right corner of the rectangle: the *max* coord.

.. doctest:: box_demo
    :hide:

    >>> from benker.box import Box
    >>> from benker.coord import Coord

To instantiate a :class:`~benker.box.Box`, you can do:

.. doctest:: box_demo

    >>> b1 = Box(Coord(5, 6), Coord(7, 8))
    >>> b2 = Box(Coord(5, 6))
    >>> b3 = Box(1, 2, 2, 3)
    >>> b4 = Box(1, 2)
    >>> b5 = Box(b1)

*Box* objects have a string representation à la Excel:

.. doctest:: box_demo

    >>> for box in b1, b2, b3, b4, b5:
    ...     print(box)
    E6:G8
    E6
    A2:B3
    A2
    E6:G8

You can calculate the *width* and *height* of boxes:

.. doctest:: box_demo

    >>> b1 = Box(Coord(5, 6), Coord(6, 8))
    >>> b1.width, b1.height
    (2, 3)

    >>> b2 = Box(Coord(5, 6))
    >>> b2.width, b2.height
    (1, 1)

You can determine if a *Coord* is included in a *Box*:

.. doctest:: box_demo

    >>> top_left = Coord(5, 6)
    >>> top_right = Coord(6, 6)
    >>> bottom_left = Coord(5, 8)
    >>> bottom_right = Coord(6, 8)

    >>> b1 = Box(top_left, bottom_right)

    >>> top_left in b1
    True
    >>> top_right in b1
    True
    >>> bottom_left in b1
    True
    >>> bottom_right in b1
    True

    >>> Coord(7, 6) in b1
    False

    >>> (5, 7) in b1
    True

You can determine if two boxes intersect each other, or are disjoints:

.. doctest:: box_demo

    >>> b1 = Box(Coord(5, 6), Coord(6, 8))
    >>> b2 = Box(Coord(6, 6), Coord(6, 7))
    >>> b3 = Box(Coord(7, 6), Coord(7, 8))
    >>> b2.intersect(b3)
    False
    >>> b1.isdisjoint(b2)
    False
    >>> b2.isdisjoint(b1)
    False
    >>> b1.isdisjoint(b3)
    True
    >>> b3.isdisjoint(b1)
    True

"""
import collections
import functools

from benker.coord import Coord
from benker.size import Size

BoxTuple = collections.namedtuple('BoxTuple', ['min', 'max'])


@functools.total_ordering
class Box(BoxTuple):
    """
    A *Box* is a rectangular area defined by two coordinates:

    - the top-left corner of the rectangle: the *min* coord,
    - the bottom-right corner of the rectangle: the *max* coord.

    Usage:

    .. doctest:: box_demo

        >>> from benker.box import Box

        >>> box = Box(1, 1, 5, 3)
        >>> box
        Box(min=Coord(x=1, y=1), max=Coord(x=5, y=3))

    """
    __slots__ = ()

    def __new__(cls, *args):
        """
        Construct a new *Box*.

        :param args:
            Arguments could be:

            - top-left and bottom-right coordinates of the box: Coord(*min_coord*, *max_coord*);
            - top-left coordinates and box size: Coord(*min_coord*, *max_coord*), Size(*width*, *height*);
            - top-left coordinates of the box: Coord(*min_coord*, *max_coord*),
              assuming box size is (1, 1),
            - coordinates of the box: *min_x*, *min_y*, *max_x*, *max_y*;
            - coordinates of the top-left coord: *min_x*, *min_y*,
              assuming box size is (1, 1);
            - another box.

        :return: The new *Box*.

        :raises TypeError:
            if the arguments are of incompatible types.

        :raises ValueError:
        """
        types = tuple(map(type, args))
        if types == (Coord, Coord):
            min_x, min_y = args[0]
            max_x, max_y = args[1]
        elif types == (Coord, Size):
            min_x, min_y = args[0]
            max_x, max_y = args[0] + args[1] - 1
        elif types == (Coord,):
            min_x, min_y = args[0]
            max_x, max_y = min_x, min_y
        elif types == (int, int, int, int):
            min_x, min_y, max_x, max_y = args
        elif types == (int, int):
            min_x, min_y = args
            max_x, max_y = min_x, min_y
        elif types == (cls,):
            # no duplicate
            return args[0]
        else:
            raise TypeError(repr(types))
        if 0 < min_x <= max_x and 0 < min_y <= max_y:
            min_coord = Coord(min_x, min_y)
            max_coord = Coord(max_x, max_y)
            # noinspection PyArgumentList
            return super(Box, cls).__new__(cls, min_coord, max_coord)
        raise ValueError(*args)

    def __str__(self):
        if (self.width, self.height) == (1, 1):
            return str(self.min)
        return str(self.min) + ':' + str(self.max)

    def __repr__(self):
        return super(Box, self).__repr__().replace('BoxTuple', 'Box')

    @property
    def width(self):
        # type: () -> int
        return self.max.x - self.min.x + 1

    @property
    def height(self):
        # type: () -> int
        return self.max.y - self.min.y + 1

    @property
    def size(self):
        return Size(self.width, self.height)

    def transform(self, coord=None, size=None):
        min_coord = self.min if coord is None else Coord.from_value(coord)
        size = self.size if size is None else Size.from_value(size)
        max_coord = min_coord + size - 1
        return Box(min_coord, max_coord)

    def move_to(self, coord):
        return self.transform(coord=coord)

    def resize(self, size):
        return self.transform(size=size)

    def __contains__(self, coord):
        coord_type = type(coord)
        if coord_type is Coord:
            return self.min.x <= coord.x <= self.max.x and self.min.y <= coord.y <= self.max.y
        elif coord_type is tuple and tuple(map(type, coord)) == (int, int):
            return self.min.x <= coord[0] <= self.max.x and self.min.y <= coord[1] <= self.max.y
        elif coord_type is Box:
            return coord.min in self and coord.max in self
        raise TypeError(repr(coord_type))

    def intersect(self, that):
        # type: (Box) -> bool
        return ((self.min in that or self.max in that) or
                (that.min in self or that.max in self))

    def isdisjoint(self, that):
        # type: (Box) -> bool
        return not self.intersect(that)

    def union(self, *others):
        """
        Return the union of *self* and all the *boxes*.

        Usage:

        .. doctest:: box_demo

            >>> from benker.box import Box
            >>> from benker.coord import Coord

            >>> b1 = Box(Coord(3, 2), Coord(6, 4))
            >>> b2 = Box(Coord(4, 3), Coord(5, 7))
            >>> b1.union(b2)
            Box(min=Coord(x=3, y=2), max=Coord(x=6, y=7))

            >>> b1 | b2
            Box(min=Coord(x=3, y=2), max=Coord(x=6, y=7))

        :param others: collections of boxes

        :return: The bounding box of all the boxes.
        """
        boxes = (self,) + others
        min_list = tuple(box.min for box in boxes)
        min_coord = Coord(min(coord.x for coord in min_list), min(coord.y for coord in min_list))
        max_list = tuple(box.max for box in boxes)
        max_coord = Coord(max(coord.x for coord in max_list), max(coord.y for coord in max_list))
        bounding_box = Box(min_coord, max_coord)
        assert all(box in bounding_box for box in boxes)
        return bounding_box

    __or__ = union

    def intersection(self, *others):
        """
        Return the intersection of *self* and all the *boxes*.

        Usage:

        .. doctest:: box_demo

            >>> from benker.box import Box
            >>> from benker.coord import Coord

            >>> b1 = Box(Coord(3, 2), Coord(6, 4))
            >>> b2 = Box(Coord(4, 3), Coord(5, 7))
            >>> b1.intersection(b2)
            Box(min=Coord(x=4, y=3), max=Coord(x=5, y=4))

            >>> b1 & b2
            Box(min=Coord(x=4, y=3), max=Coord(x=5, y=4))

        :param others: collections of boxes

        :return: The inner box of all the boxes.

        :raises ValueError: if the two boxes are disjoint.
        """
        boxes = (self,) + others
        min_list = tuple(box.min for box in boxes)
        min_coord = Coord(max(coord.x for coord in min_list), max(coord.y for coord in min_list))
        max_list = tuple(box.max for box in boxes)
        max_coord = Coord(min(coord.x for coord in max_list), min(coord.y for coord in max_list))
        try:
            return Box(min_coord, max_coord)
        except ValueError:
            # the two boxes are disjoint
            raise ValueError(boxes)

    __and__ = intersection

    # total ordering based on coordinates (*y* first, then *x*).
    # This or ordering can be used to sort boxes by rows and columns

    def __lt__(self, other):
        """
        Compare two boxes.

        Usage::

        .. doctest:: box_demo

            >>> from benker.box import Box

            >>> b1 = Box(Coord(3, 2), Coord(6, 4))
            >>> b1 < b1
            False
            >>> b1 < Box(Coord(3, 2), Coord(6, 5))
            True
            >>> b1 < Box(Coord(3, 2), Coord(7, 4))
            True
            >>> b1 < Box(Coord(4, 2), Coord(6, 4))
            True
            >>> b1 < Box(Coord(3, 3), Coord(6, 4))
            True

        :param other: other box

        :return: ``True`` if *self* < *other*
        """
        if isinstance(other, Box):
            if self.min.y == other.min.y:
                if self.min.x == other.min.x:
                    if self.max.y == other.max.y:
                        return self.max.x < other.max.x
                    else:
                        return self.max.y < other.max.y
                else:
                    return self.min.x < other.min.x
            else:
                return self.min.y < other.min.y
        return NotImplemented



# coding: utf-8
"""
Office Open XML to Formex4 converter
====================================
"""
from benker.builders.formex4 import Formex4Builder
from benker.converters.base_converter import BaseConverter
from benker.parsers.ooxml import OoxmlParser


class Ooxml2Formex4Converter(BaseConverter):
    """
    Office Open XML to Formex4 converter
    """
    parser_cls = OoxmlParser
    builder_cls = Formex4Builder


def convert_ooxml2formex4(src_xml, dst_xml, **options):
    """
    Convert Office Open XML (OOXML) tables to Formex4 tables.

    :param str src_xml:
        Source path of the XML file to convert.

        This must be an XML file, for instance, if you want to convert a Word
        document (``.docx``), you first need to unzip the ``.docx`` file, and
        then, run this function on the file ``word/document.xml``.
        You can also use the *styles_path* option to parse and use the table
        styles defined in the file ``word/styles.xml``.

    :param str dst_xml:
        Destination path of the XML file to produce.

    :keyword options: Extra conversion options.
        See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
        to have a list of all possible options.
    """
    converter = Ooxml2Formex4Converter()
    converter.convert_file(src_xml, dst_xml, **options)



# coding: utf-8
"""
Base Builder
============

Base class of Builders.
"""


class BaseBuilder(object):
    """
    Base class of Builders.
    """

    def __init__(self, **options):
        """
        Construct a builder.

        :param options: building options.
        """
        self.options = options

    def generate_table_tree(self, table):
        """
        Build the XML table from the Table instance.

        :type  table: benker.table.Table
        :param table: Table

        :return: Table tree
        """
        raise NotImplementedError

    def finalize_tree(self, tree):
        """
        Give the opportunity to finalize the resulting tree structure.

        :type  tree: etree._ElementTree
        :param tree: The resulting tree.

        .. versionadded:: 0.4.0
        """



# coding: utf-8
"""
Table
=====

Generic table structure which simplify the conversion from docx table format to CALS or HTML tables.
"""
import re

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

from benker.cell import Cell
from benker.coord import Coord
from benker.grid import Grid
from benker.styled import Styled


# noinspection PyUnresolvedReferences
class TableViewList(object):
    """
    This class defined a (simplified) list of views.

    Short demonstration:

    .. doctest:: table_demo

        >>> from benker.cell import Cell
        >>> from benker.table import Table
        >>> from benker.table import ColView
        >>> from benker.table import RowView
        >>> from benker.table import TableViewList

        >>> red = Cell('red', x=1, y=1, height=2)
        >>> pink = Cell('pink', x=2, y=1, width=2)
        >>> blue = Cell('blue', x=2, y=2)
        >>> table = Table([red, pink, blue])
        >>> print(table)
        +-----------+-----------------------+
        |    red    |   pink                |
        |           +-----------+-----------+
        |           |   blue    |           |
        +-----------+-----------+-----------+

        >>> cols = TableViewList(table, ColView)
        >>> len(cols)
        3
        >>> rows = TableViewList(table, RowView)
        >>> len(rows)
        2

        >>> for pos, col in enumerate(cols, 1):
        ...     print("col #{pos}: {col}".format(pos=pos, col=str(col)))
        col #1: [red]
        col #2: [pink, blue]
        col #3: []

        >>> cols[3].insert_cell("yellow")
        >>> print(table)
        +-----------+-----------------------+
        |    red    |   pink                |
        |           +-----------+-----------+
        |           |   blue    |  yellow   |
        +-----------+-----------+-----------+
    """

    def __init__(self, table, view_cls):
        """
        Construct a list of view attached to the given *table*.

        :type  table: benker.table.Table
        :param table: Reference to the table.

        :type  view_cls: type
        :param view_cls:
            A view class, either :class:`~benker.table.RowView` or :class:`~benker.table.ColView`.
        """
        self.table = table
        self.view_cls = view_cls
        self.views = []
        self.refresh_all()

    def adopt_cell(self, cell):
        """
        Adopt a new cell in the views.

        :type  cell: benker.cell.Cell
        :param cell: New cell to adopt
        """
        views = self.views
        box = cell.box
        if self.view_cls is RowView:
            self._fit_to_size(box.max.y, truncate=False)
            for pos in range(box.min.y, box.max.y + 1):
                views[pos - 1].adopt_cell(cell)
        elif self.view_cls is ColView:
            self._fit_to_size(box.max.x, truncate=False)
            for pos in range(box.min.x, box.max.x + 1):
                views[pos - 1].adopt_cell(cell)
        else:
            raise TypeError(repr(self.view_cls))

    def refresh_all(self):
        """
        Cleanup and refresh all the views, taking into account the cells
        which are in the table grid.
        """
        views = self.views
        bounding_box = self.table.bounding_box
        if bounding_box is None:
            views[:] = []
            return

        if self.view_cls is RowView:
            self._fit_to_size(bounding_box.max.y)
        elif self.view_cls is ColView:
            self._fit_to_size(bounding_box.max.x)
        else:
            raise TypeError(repr(self.view_cls))

        for view in views:
            view.owned_cells[:] = []
            view.caught_cells[:] = []

        for cell in self.table:
            self.adopt_cell(cell)

    def _fit_to_size(self, size, truncate=True):
        views = self.views
        if truncate:
            del views[size:]
        for index in range(len(views), size):
            views.append(self.view_cls(self.table, index + 1))

    def __getitem__(self, pos):
        self._fit_to_size(pos, truncate=False)
        return self.views[pos - 1]

    def __len__(self):
        return len(self.views)

    def __iter__(self):
        return iter(self.views)


def _camel_to_snake(camel_name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel_name)
    snake_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
    return snake_name


class ViewsProperty(object):
    """
    Descriptor used to define the rows/cols properties
    in the class :class:`~benker.table.Table`.
    """

    def __init__(self, view_cls):
        """
        Construct a "Views" property which handles a list of items which type is *view_cls*.

        :type  view_cls: type
        :param view_cls:
            A view class, either :class:`~benker.table.RowView` or :class:`~benker.table.ColView`.
        """
        self.view_cls = view_cls
        # CamelCase to snake_case:
        cls_module = self.view_cls.__module__
        cls_name = self.view_cls.__name__
        self.attr_name = "_{}".format(_camel_to_snake(cls_name))
        self.__doc__ = "List of views of type :class:`~{0}.{1}`".format(cls_module, cls_name)

    def __get__(self, table, owner):
        if table is None:
            return self
        if not hasattr(table, self.attr_name):
            setattr(table, self.attr_name, TableViewList(table, self.view_cls))
        return getattr(table, self.attr_name)


class TableView(Styled):
    """
    Base class of :class:`~benker.table.RowView` and :class:`~benker.table.ColView`
    used to create a view on the table cells for a given row or column.

    See also: :class:`~benker.table.TableViewList`
    """
    __slots__ = ('_table', '_pos', '_owned_cells', '_caught_cells')

    def __init__(self, table, pos, styles=None, nature=None):
        """
        Initialize a row/column view.

        :type  table: benker.table.Table
        :param table: Reference to the table which contains the cells to "view".

        :param int pos:
            Row or column pos (1-based) used to select cells.

        :type  styles: typing.Dict[str, str]
        :param styles:
            Dictionary of key-value pairs, where *keys* are the style names.

        :param str nature:
            Cell *nature* used to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.

        .. versionchanged:: 0.4.2
           The default value of *nature* is ``None`` (instead of "body").
        """
        super(TableView, self).__init__(styles, nature)
        self._table = table
        self._pos = pos
        self._owned_cells = []
        self._caught_cells = []

    def __str__(self):
        items = ", ".join(str(cell) for cell in self._owned_cells)
        return "[{0}]".format(items)

    @property
    def table(self):
        """ Non-mutable reference to the table (instance of :class:`~benker.table.Table`). """
        return self._table

    @property
    def owned_cells(self):
        """ List of cells owned by this view. """
        return self._owned_cells

    @property
    def caught_cells(self):
        """ List of cells caught (intercepted) by this view. """
        return self._caught_cells

    def can_own(self, cell):
        """
        Check if a cell can be stored it that view.

        :type  cell: benker.cell.Cell
        :param cell: table cell
        :return: ``True`` if the cell belong to this view.
        """
        raise NotImplementedError

    def can_catch(self, cell):
        """
        Check if a cell can be caught by that view.

        :type  cell: benker.cell.Cell
        :param cell: table cell
        :return: ``True`` if the cell intercept to this view.
        """
        raise NotImplementedError

    def adopt_cell(self, cell):
        """
        Event handler called by the system when
        a cell is about to be inserted in the table.
        """
        if self.can_own(cell):
            self._owned_cells.append(cell)
        if self.can_catch(cell):
            self._caught_cells.append(cell)


class RowView(TableView):
    """
    A view on the table cells for a given row.
    """

    def can_own(self, cell):
        return cell.min.y == self._pos

    def can_catch(self, cell):
        return Coord(cell.min.x, self._pos) in cell

    @property
    def row_pos(self):
        """ Row position in the table (1-based). """
        return self._pos

    def insert_cell(self, content, styles=None, nature=None, width=1, height=1):
        """
        Insert a new cell in the row at the next free position, or at the end.

        :param content:
            User-defined cell content. It can be of any type: ``None``,
            :class:`str`, :class:`int`, :class:`float`, a container (:class:`list`),
            a XML element, etc. The same content can be shared by several cells, it's
            your own responsibility to handle the copy (or deep copy) of the *content*
            reference when needed.

        :type  styles: typing.Dict[str, str]
        :param styles:
            User-defined cell styles: a dictionary of key-value pairs.
            This values are useful to store some HTML-like styles (border-style,
            border-width, border-color, vertical-align, text-align, etc.).
            Of course, we are not tied to the HTML-like styles, you can use your own
            list of styles.

        :type nature: str
        :ivar nature: a way to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.
            If set to ``None``, the cell nature is inherited from the row nature.

        :param int width:
            Width of the cell (columns spanning), default to 1.

        :param int height:
            Height of the cell (rows spanning), default to 1.

        .. versionchanged:: 0.4.2
           The *nature* of a cell is inherited from its parent's row.
        """
        caught_cells = self._caught_cells
        y = self._pos
        if caught_cells:
            boxes = tuple(cell.box for cell in caught_cells)
            bounding_box = boxes[0].union(*boxes[1:])
            for x in range(1, bounding_box.max.x + 1):
                if all((x, y) not in box for box in boxes):
                    break
            else:
                x = bounding_box.max.x + 1
        else:
            x = 1
        nature = self.nature if nature is None else nature
        cell = Cell(content, styles=styles, nature=nature,
                    x=x, y=y, width=width, height=height)
        self._table[cell.min] = cell


class ColView(TableView):
    """
    A view on the table cells for a given column.
    """

    def can_own(self, cell):
        return cell.min.x == self._pos

    def can_catch(self, cell):
        return Coord(self._pos, cell.min.y) in cell

    @property
    def col_pos(self):
        """ Column position in the table (1-based). """
        return self._pos

    def insert_cell(self, content, styles=None, nature=None, width=1, height=1):
        """
        Insert a new cell in the column at the next free position, or at the end.

        :param content:
            User-defined cell content. It can be of any type: ``None``,
            :class:`str`, :class:`int`, :class:`float`, a container (:class:`list`),
            a XML element, etc. The same content can be shared by several cells, it's
            your own responsibility to handle the copy (or deep copy) of the *content*
            reference when needed.

        :type  styles: typing.Dict[str, str]
        :param styles:
            User-defined cell styles: a dictionary of key-value pairs.
            This values are useful to store some HTML-like styles (border-style,
            border-width, border-color, vertical-align, text-align, etc.).
            Of course, we are not tied to the HTML-like styles, you can use your own
            list of styles.

        :type nature: str
        :ivar nature: a way to distinguish the body cells, from the header and the footer.
            The default value is ``None``, but you can use "body", "header", "footer" or whatever
            is suitable for your needs.
            If set to ``None``, the cell nature is inherited from the column nature.

        :param int width:
            Width of the cell (columns spanning), default to 1.

        :param int height:
            Height of the cell (rows spanning), default to 1.

        .. versionchanged:: 0.4.2
           The *nature* of a cell is inherited from its parent's column.
        """
        caught_cells = self._caught_cells
        x = self._pos
        if caught_cells:
            boxes = tuple(cell.box for cell in caught_cells)
            bounding_box = boxes[0].union(*boxes[1:])
            for y in range(1, bounding_box.max.y + 1):
                if all((x, y) not in box for box in boxes):
                    break
            else:
                y = bounding_box.max.y + 1
        else:
            y = 1
        nature = self.nature if nature is None else nature
        cell = Cell(content, styles=styles, nature=nature,
                    x=x, y=y, width=width, height=height)
        self._table[cell.min] = cell


class Table(Styled, MutableMapping):
    """
    Table data structure used to simplify conversion to CALS or HTML.

    Short demonstration:

    .. doctest:: table_demo

        >>> from benker.cell import Cell
        >>> from benker.table import Table

        >>> table = Table(styles={'frame': 'all'})

        >>> table[(1, 1)] = Cell("one")
        >>> table.rows[1].insert_cell("two")

        >>> table[(2, 1)]
        <Cell('two', styles={}, nature=None, x=2, y=1, width=1, height=1)>

        >>> table.cols[1].insert_cell("alpha")
        >>> table.cols[2].insert_cell("beta")
        >>> (1, 2) in table
        True

        >>> del table[(1, 2)]
        >>> (1, 2) in table
        False

        >>> len(table)
        3

        >>> for cell in table:
        ...     print(cell)
        one
        two
        beta

        >>> for row in table.rows:
        ...     print(row)
        [one, two]
        [beta]

        >>> table.merge((1, 2), (2, 2))
        >>> print(table)
        +-----------+-----------+
        |    one    |    two    |
        +-----------------------+
        |   beta                |
        +-----------------------+

        >>> table.expand((1, 1), width=3)
        >>> print(table)
        +-----------------------------------------------+
        |              onetwo                           |
        +-----------------------+-----------+-----------+
        |   beta                |           |           |
        +-----------------------+-----------+-----------+

    :type styles: typing.Dict[str, str]
    :ivar styles:
        User-defined table styles: a dictionary of key-value pairs.
        This values are useful to store some HTML-like styles (border-style,
        border-width, border-color, vertical-align, text-align, etc.).
        Of course, we are not tied to the HTML-like styles, you can use your own
        list of styles.

        .. note::

            The style dictionary is always copied: in other words, key-value pairs
            are copied but a shallow copy is done for the values (in general, it
            is not a problem if you use non-mutable values like :class:`str`).

    :type nature: str
    :ivar nature:
        A table can have a nature: a way to distinguish the body cells,
        from the header and the footer. The default value is ``None``, but you can
        use "body", "header", "footer" or whatever is suitable for your needs.
        This kind of information is in general not stored in the styles,
        even if it is similar.

        .. note::

            In a :class:`~benker.grid.Grid`, the :ref:`merging <benker__grid__merging>`
            of two natures is done by keeping the first nature and
            dropping the second one. In other words, the resulting nature is
            the group of the most top-left nature of the merged cells.
    """

    rows = ViewsProperty(RowView)
    cols = ViewsProperty(ColView)

    def __init__(self, cells=None, styles=None, nature=None):
        """
        Construct a table object from a collection of cells and a dictionary of styles.

        :type  cells: typing.Iterable[benker.cell.Cell]
        :param cells: Collection of cells.

        :type  styles: typing.Dict[str, str]
        :param styles:
            Dictionary of key-value pairs, where *keys* are the style names.

        :type nature: str
        :ivar nature:
            User-defined string value.
            Table *nature* is similar to HTML ``@class`` attribute,
            you can use it do identify the styles to apply to your table.
        """
        super(Table, self).__init__(styles, nature)
        self._grid = Grid(cells)

    def __str__(self):
        return str(self._grid)

    @property
    def bounding_box(self):
        """
        Bounding box of the table (``None`` if the table is empty).

        :rtype: benker.box.Box
        :return: The bounding box.
        """
        return self._grid.bounding_box

    def _refresh_views(self, cell=None):
        """
        Refresh all the rows and column views.
        """
        if cell:
            self.rows.adopt_cell(cell)
            self.cols.adopt_cell(cell)
        else:
            self.rows.refresh_all()
            self.cols.refresh_all()

    def __contains__(self, coord):
        """
        Check if the table contains a cell (a owned cell) at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).

        :return: ``True`` if the table contains a cell at the given
            coordinates, else ``False``.
        """
        return coord in self._grid

    def __setitem__(self, coord, cell):
        """
        Insert a cell in the table at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).

        :type  cell: benker.cell.Cell
        :param cell: Cell to insert.
        """
        self._grid[coord] = cell
        # get the cell with its new coordinates:
        cell = self._grid[coord]
        self._refresh_views(cell)

    def __delitem__(self, coord):
        """
        Delete a cell from the table at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).
        """
        del self._grid[coord]
        self._refresh_views()

    def __getitem__(self, coord):
        """
        Get a cell at the given coordinates.

        :type  coord: Coord or tuple(int, int)
        :param coord: Coordinates in the tables (1-indexed).

        :rtype: benker.cell.Cell
        :return: The cell.
        """
        return self._grid[coord]

    def __len__(self):
        return len(self._grid)

    def __iter__(self):
        return iter(self._grid)

    def merge(self, start, end, content_appender=None):
        self._grid.merge(start, end, content_appender=content_appender)
        self._refresh_views()

    def expand(self, coord, width=0, height=0, content_appender=None):
        self._grid.expand(coord, width=width, height=height, content_appender=content_appender)
        self._refresh_views()



# coding: utf-8
"""
Base Converter
==============

Bas class of all converters.
"""
from benker.builders.base_builder import BaseBuilder
from benker.parsers.base_parser import BaseParser


class BaseConverter(object):
    """
    Bas class of all converters.
    """

    parser_cls = BaseParser
    builder_cls = BaseBuilder

    def convert_file(self, src_xml, dst_xml, **options):
        """
        Convert the tables from one format to another.

        :param str src_xml:
            Source path of the XML file to convert.

        :param str dst_xml:
            Destination path of the XML file to produce.

        :keyword options:
            Dictionary of parsing/building options.

            **Common parsing options:**

            ``encoding`` (default: "utf-8"):
                XML encoding of the destination file.

            **OOXML parser options:**

            ``styles_path`` (default: ``None``):
                Path to the stylesheet to use to resole table styles.
                In an uncompressed ``.docx`` tree structure, the stylesheet path
                is ``word/styles.xml``.

            **Common builder options:** *(none)*

            **CALS builder options:**

            ``width_unit`` (default: "mm"):
                Unit to use for column widths.
                Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.

            ``table_in_tgroup`` (default: ``False``):
                Where should we put the table properties:

                -   ``False`` to insert the attributes ``@colsep``, ``@rowsep``,
                    and ``@tabstyle`` in the ``<table>`` element,

                -   ``True`` to insert the attributes ``@colsep``, ``@rowsep``,
                    and ``@tgroupstyle`` in the ``<tgroup>`` element.

            **Formex4 builder options:** *(none)*

        """
        builder = self.builder_cls(**options)
        parser = self.parser_cls(builder, **options)
        parser.parse_file(src_xml, dst_xml)



# coding: utf-8
"""
Base Parser
===========

Base class of parsers.
"""
from lxml import etree


class BaseParser(object):
    """
    Abstract base class of the parsers classes.
    """

    class _State(object):
        """
        Parsing state for the converter (internal usage).

        .. versionadded:: 0.4.4
        """

        def __init__(self):
            self.col_pos = 0
            self.col = None
            self.row_pos = 0
            self.row = None
            self.table = None

        reset = __init__

        def next_col(self):
            self.col_pos += 1
            self.col = None

        def next_row(self):
            self.col_pos = 0
            self.col = None
            self.row_pos += 1
            self.row = None

    def __init__(self, builder, encoding="utf-8", **options):
        """
        Construct a base builder.

        :type  builder: benker.builders.base_builder.BaseBuilder
        :param builder:
            Builder used by this parser to instantiate :class:`~benker.table.Table` objects.

        :param str encoding:
            XML encoding of the destination file (default: "utf-8").

        :keyword options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.
        """
        self.builder = builder
        self.encoding = encoding
        self.options = options
        self._state = self._State()

    def parse_file(self, src_xml, dst_xml):
        """
        Parse and convert the tables from one format to another.

        :param str src_xml:
            Source path of the XML file to convert.

        :param str dst_xml:
            Destination path of the XML file to produce.
        """

        tree = etree.parse(src_xml)
        self.transform_tables(tree)
        self.builder.finalize_tree(tree)
        tree.write(dst_xml, encoding=self.encoding, pretty_print=False)

    def transform_tables(self, tree):
        raise NotImplementedError


def value_of(element, xpath, namespaces=None, default=None):
    """
    Take the first value of a xpath evaluation.

    :type  element: etree._Element
    :param element: Root element used to evaluate the xpath expression.

    :param str xpath: xpath expression.
        This expression will be evaluated using the *namespaces* namespaces.

    :type  namespaces: dict[str, str]
    :param namespaces:
        Namespace map to use for the xpath evaluation.

    :param default: default value used if the xpath evaluation returns no result.

    :return: the first result or the *default* value.
    """
    if element is None:
        return default
    nodes = element.xpath(xpath, namespaces=namespaces)
    return nodes[0] if nodes else default



# coding: utf-8
"""
Alphabet
========

Utility functions to convert integer into a base-26 "number", and vis versa.
"""

try:
    from string import uppercase as ascii_uppercase
except ImportError:
    from string import ascii_uppercase


def int_to_alphabet(value, alphabet=ascii_uppercase):
    """
    Convert a non-nul integer into a base-26 "number" using uppercase ASCII letters.

    Usage:

    .. doctest:: alphabet_demo

        >>> from benker.alphabet import int_to_alphabet

        >>> int_to_alphabet(1)
        'A'
        >>> int_to_alphabet(2)
        'B'
        >>> int_to_alphabet(26)
        'Z'
        >>> int_to_alphabet(27)
        'AA'
        >>> int_to_alphabet(28)
        'AB'
        >>> int_to_alphabet(52)
        'AZ'
        >>> int_to_alphabet(53)
        'BA'
        >>> int_to_alphabet(18278)
        'ZZZ'
        >>> int_to_alphabet(-5)
        Traceback (most recent call last):
            ...
        ValueError: -5

    :param int value: value to convert
    :param alphabet: alphabet to use for the conversion.
    :return: string representing this "number" in the base-26.
    """
    if value == 0:
        return ""
    elif value < 0:
        raise ValueError(value)
    size = len(alphabet)
    q, r = divmod(value - 1, size)
    letters = [alphabet[r]]
    while q:
        q, r = divmod(q - 1, size)
        letters.insert(0, alphabet[r])
    return "".join(letters)


def alphabet_to_int(letters, alphabet=ascii_uppercase):
    """
    Convert a base-26 "number" using uppercase ASCII letters into an integer.

    .. doctest:: alphabet_demo

        >>> from benker.alphabet import alphabet_to_int

        >>> alphabet_to_int("A")
        1
        >>> alphabet_to_int("B")
        2
        >>> alphabet_to_int("AA")
        27
        >>> alphabet_to_int("AB")
        28
        >>> alphabet_to_int("ZZZ")
        18278
        >>> alphabet_to_int("")
        0
        >>> alphabet_to_int("AA@")
        Traceback (most recent call last):
            ...
        ValueError: AA@

    :param letters: string representing a "number" in the base-26.
    :param alphabet: alphabet to use for the conversion.
    :return: Integer value of the "number".
    """
    value = 0
    size = len(alphabet)
    try:
        for letter in letters:
            value *= size
            value += alphabet.index(letter) + 1
        return value
    except ValueError:
        raise ValueError(letters)
